/***************************************************************************
 *   Copyright (C) 2009 by Ben Wolf   *
 *   bswolf@cs.cmu.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "graphics.h"
#include "direction.h"
#include "Enemy.h"
#include <algorithm>

Enemy::Enemy(int x, int y, int dir, list<point> path, int vel,
		int rate, int offset,
		SDL_Surface* image, void (*laser_callback)(Laser)) :
			path(path), vel(vel), rate(rate), count(offset),
			Ship(dir, x, y, image, laser_callback) {

}

void Enemy::move() {
	if (state != ALIVE) return;

	count++;
	if (count >= rate) {
		this->fireLaser();
		count = 0;
	}
	point c = path.front();
	int dx = c.x - pos_x;
	int dy = c.y - pos_y;

	if (dx == 0 && dy == 0) {
		path.push_back(c);
		path.pop_front();
		c = path.front();
		dx = c.x - pos_x;
		dy = c.y - pos_y;
	}

	vel_x = (dx >= 0) ? min(dx, vel) : max(dx, -vel);
	vel_y = (dy >= 0) ? min(dy, vel) : max(dy, -vel);

	// Go towards that point at vel.
	this->Sprite::move();
}

point Enemy::shift(point a) {
	point d;
	d.x = a.x + pos_x;
	d.y = a.y + pos_y;
	return d;
}

static SDL_Surface* findForwardImage(int dir) {
	switch(dir) {
	case NORTH:	return GetImage(IMG_FORWARD_N);
	case EAST:	return GetImage(IMG_FORWARD_E);
	case SOUTH:	return GetImage(IMG_FORWARD_S);
	case WEST:	return GetImage(IMG_FORWARD_W);
	}
	return NULL;
}

#define HITBOX_P1X 24
#define HITBOX_P1Y 5
#define HITBOX_P2X 9
#define HITBOX_P2Y 44
#define HITBOX_P3X 41
#define HITBOX_P3Y 44

FShooter::FShooter(int x, int y, int dir, list<point> path, int vel, int rate,
		int offset, void (*laser_callback)(Laser)) : Enemy(x, y, dir, path,
			vel, rate, offset, findForwardImage(dir), laser_callback) {
	switch(dir) {
	case NORTH:
		u1.x = HITBOX_P1X;
		u1.y = HITBOX_P1Y;
		u2.x = HITBOX_P2X;
		u2.y = HITBOX_P2Y;
		u3.x = HITBOX_P3X;
		u3.y = HITBOX_P3Y;
		break;
	case EAST:
		u1.x = image->w - HITBOX_P1Y;
		u1.y = HITBOX_P1X;
		u2.x = image->w - HITBOX_P2Y;
		u2.y = HITBOX_P2X;
		u3.x = image->w - HITBOX_P3Y;
		u3.y = HITBOX_P3X;
		break;
	case SOUTH:
		u1.x = image->w - HITBOX_P1X;
		u1.y = image->h - HITBOX_P1Y;
		u2.x = image->w - HITBOX_P2X;
		u2.y = image->h - HITBOX_P2Y;
		u3.x = image->w - HITBOX_P3X;
		u3.y = image->h - HITBOX_P3Y;
		break;
	case WEST:
		u1.x = HITBOX_P1Y;
		u1.y = image->h - HITBOX_P1X;
		u2.x = HITBOX_P2Y;
		u2.y = image->h - HITBOX_P2X;
		u3.x = HITBOX_P3Y;
		u3.y = image->h - HITBOX_P3X;
		break;
	}
}

void FShooter::fireLaser() {
	if (state != ALIVE) return;
	point c = center_side(direction);
	Laser* laser = new Laser(c.x, c.y, direction, LASER_SPEED, LASER_LIFE,
			ENEMY);
	if (firedLaser != NULL) {
		firedLaser(*laser);
	}
}

bool FShooter::doesCollide(int x, int y) {
	point c;
	c.x = x;
	c.y = y;
	return inTriangle(c, shift(u1), shift(u2), shift(u3));
}

static SDL_Surface* findDualImage(int dir) {
	switch(dir) {
	case NORTH:
	case SOUTH:	return GetImage(IMG_DUAL_V);
	case EAST:
	case WEST:	return GetImage(IMG_DUAL_H);
	}
	return NULL;
}

#define DBOX_P1X 2
#define DBOX_P1Y 24
#define DBOX_P2X 20
#define DBOX_P2Y 2
#define DBOX_P3X 20
#define DBOX_P3Y 48
#define DBOX_CENTER 15

DShooter::DShooter(int x, int y, int dir, list<point> path, int vel, int rate,
		int offset, void (*laser_callback)(Laser)) : Enemy(x, y, dir, path,
			vel, rate, offset, findDualImage(dir), laser_callback) {
	switch(dir) {
	case NORTH:
	case SOUTH:
		u1.x = DBOX_P1X;
		u1.y = DBOX_P1Y;
		u2.x = DBOX_P2X;
		u2.y = DBOX_P2Y;
		u3.x = DBOX_P3X;
		u3.y = DBOX_P3Y;
		u4.x = image->w - DBOX_P1X;
		u4.y = DBOX_P1Y;
		u5.x = image->w - DBOX_P2X;
		u5.y = DBOX_P2Y;
		u6.x = image->w - DBOX_P3X;
		u6.y = DBOX_P3Y;
		break;
	case EAST:
	case WEST:
		u1.x = image->w - DBOX_P1Y;
		u1.y = DBOX_P1X;
		u2.x = image->w - DBOX_P2Y;
		u2.y = DBOX_P2X;
		u3.x = image->w - DBOX_P3Y;
		u3.y = DBOX_P3X;
		u4.x = image->w - DBOX_P1Y;
		u4.y = image->h - DBOX_P1X;
		u5.x = image->w - DBOX_P2Y;
		u5.y = image->h - DBOX_P2X;
		u6.x = image->w - DBOX_P3Y;
		u6.y = image->h - DBOX_P3X;
		break;
	}
}

bool DShooter::doesCollide(int x, int y) {
	point c;
	c.x = x;
	c.y = y;
	return (x >= pos_x + DBOX_CENTER && x <= pos_x + image->w - DBOX_CENTER &&
			y >= pos_y + DBOX_CENTER && y <= pos_y + image->h - DBOX_CENTER) ||
		inTriangle(c, shift(u1), shift(u2), shift(u3)) ||
		inTriangle(c, shift(u4), shift(u5), shift(u6));
}

void DShooter::fireLaser() {
	if (state != ALIVE) return;
	int d1 = direction;
	int d2 = rotate(direction, INVERSE);
	point c1 = center_side(d1);
	point c2 = center_side(d2);
	Laser* laser = new Laser(c1.x, c1.y, d1, LASER_SPEED, LASER_LIFE, ENEMY);
	Laser* laser2 = new Laser(c2.x, c2.y, d2, LASER_SPEED, LASER_LIFE, ENEMY);
	if (firedLaser != NULL) {
		firedLaser(*laser);
		firedLaser(*laser2);
	}
}
