/***************************************************************************
 *   Copyright (C) 2009 by Ben Wolf   *
 *   bswolf@cs.cmu.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "graphics.h"
#include "Laser.h"
#include "debug.h"

static SDL_Surface* findImage(int dir, enum laser_type type) {
	switch(type) {
	case PLAYER:
		switch(dir) {
		case NORTH: return GetImage(IMG_LASER_N);
		case EAST: return GetImage(IMG_LASER_E);
		case SOUTH: return GetImage(IMG_LASER_S);
		case WEST: return GetImage(IMG_LASER_W);
		default: return NULL;
		}
	case ENEMY:
		switch(dir) {
		case NORTH:
		case SOUTH: return GetImage(IMG_ENEMY_LASER_V);
		case EAST:
		case WEST:  return GetImage(IMG_ENEMY_LASER_H);
		default: return NULL;
		}
	}
}

Laser::Laser(int x, int y, int dir, int vel, int life, enum laser_type type)
		: dir(dir), vel(vel), vanishing(APPEARING), vanish_x(0), vanish_y(0),
		type(type), life(life), Sprite(x, y, findImage(dir, type)) {
	switch(dir) {
	case NORTH: setYVelocity(-vel);
				setXVelocity(0);
				pos_x -= image->w / 2;
				appear_x = pos_x;
				appear_y = pos_y;
				break;
	case EAST:	setXVelocity(vel);
				setYVelocity(0);
				pos_y -= image->h / 2;
				pos_x -= image->w;
				appear_x = pos_x + image->w;
				appear_y = pos_y;
				break;
	case SOUTH: setYVelocity(vel);
				setXVelocity(0);
				pos_x -= image->w / 2;
				pos_y -= image->h;
				appear_x = pos_x;
				appear_y = pos_y + image->h;
				break;
	case WEST:	setXVelocity(-vel);
				setYVelocity(0);
				pos_y -= image->h / 2;
				appear_x = pos_x;
				appear_y = pos_y;
				break;
	}
}

Laser::Laser(point c, int dir, int vel, int life, enum laser_type type):
		dir(dir), vel(vel), vanishing(APPEARING), vanish_x(0), vanish_y(0),
		type(type), life(life), Sprite(c.x, c.y, findImage(dir, type)) {
	int x = c.x,
		y = c.y;
	/* Reverse engineer where (pos_x, pos_y) should be based on
	 * where we want the hot point.
	 */
	switch(dir) {
	case NORTH:	x -= this->image->w / 2;
				y -= HP_DIST_FROM_TIP + 1;
				setYVelocity(-vel);
				setXVelocity(0);
				appear_x = x;
				appear_y = y;
				break;
	case EAST:	x -= this->image->w - HP_DIST_FROM_TIP - 1;
				y -= this->image->h / 2;
				setXVelocity(vel);
				setYVelocity(0);
				appear_x = x + image->w;
				appear_y = y;
				break;
	case SOUTH:	x -= this->image->w / 2;
				y -= this->image->h - HP_DIST_FROM_TIP - 1;
				setYVelocity(vel);
				setXVelocity(0);
				appear_x = x;
				appear_y = y + image->h;
				break;
	case WEST:	x -= HP_DIST_FROM_TIP + 1;
				y -= this->image->h / 2;
				setXVelocity(-vel);
				setYVelocity(0);
				appear_x = x;
				appear_y = y;
				break;
	}
	pos_x = x;
	pos_y = y;
}

point Laser::getHotPoint() {
	point c;
	if (vanishing == PRESENT || vanishing == APPEARING) {
		switch(dir) {
		case NORTH: c.x = this->pos_x + this->image->w / 2;
					c.y = this->pos_y + HP_DIST_FROM_TIP;
					break;
		case EAST:	c.x = this->pos_x + this->image->w - HP_DIST_FROM_TIP;
					c.y = this->pos_y + this->image->h / 2;
					break;
		case SOUTH:	c.x = this->pos_x + this->image->w / 2;
					c.y = this->pos_y + this->image->h - HP_DIST_FROM_TIP;
					break;
		case WEST:	c.x = this->pos_x + HP_DIST_FROM_TIP;
					c.y = this->pos_y + this->image->h / 2;
					break;
		}
	} else {
		c.x = -1;
		c.y = -1;
	}
	return c;
}

int Laser::getVelocity() {
	return vel;
}

int Laser::getDirection() {
	return dir;
}

int Laser::getLife() {
	return life;	
}

enum laser_type Laser::getType() {
	return type;
}

void Laser::vanish() {
	if (this->vanishing == PRESENT) {
		this->vanishing = VANISHING;
	} else if (this->vanishing == APPEARING) {
		this->vanishing = TRANSIT;
	} else {
		return;
	}
	this->vanish_x = pos_x;
	this->vanish_y = pos_y;
	if (dir == EAST) {
		vanish_x += image->w;
	} else if (dir == SOUTH) {
		vanish_y += image->h;
	}
}

bool Laser::isVanishing() {
	return vanishing == VANISHING || vanishing == TRANSIT;
}

bool Laser::vanished() {
	return vanishing == VANISHED;
}

void Laser::draw() {
	switch(vanishing) {
	case TRANSIT:
		/* In this case, the laser is both appearing and disappearing.
		 * Once the laser is done appearing, it becomes VANISHING.
		 */
		switch(dir) {
		case NORTH:
			if (appear_y - pos_y <= image->h) {
				DrawPartialImage(image, vanish_x, vanish_y, 0, vanish_y - pos_y,
					image->w, appear_y - vanish_y);
			} else {
				vanishing = VANISHING;
			}
			break;
		case EAST:
			if (appear_x - pos_x >= 0) {
				DrawPartialImage(image, appear_x, appear_y, appear_x - pos_x, 0,
					vanish_x - appear_x, image->h);
			} else {
				vanishing = VANISHING;
			}
			break;
		case SOUTH:
			if (appear_y - pos_y >= 0) {
				DrawPartialImage(image, appear_x, appear_y, 0,
					appear_y - pos_y, image->w, vanish_y - appear_y);
			} else {
				vanishing = VANISHING;
			}
			break;
		case WEST:
			if (appear_x - pos_x <= image->w) {
				DrawPartialImage(image, vanish_x, vanish_y, vanish_x - pos_x, 0,
					appear_x - vanish_x, image->h);
			} else {
				vanishing = VANISHING;
			}
			break;
		}
		if (vanishing != VANISHING) break;
		/* Otherwise, we just switched to VANISHING but we didn't draw it,
		 * so fall through to the VANISHING case.
		 */
	case VANISHING:
		/* Keeping in mind (pos_x, pos_y) is always the upper left corner
		 * of the laser, so it is also left of or above (vanish_x, vanish_y)
		 */
		switch(dir) {
		case NORTH:
			if (vanish_y - pos_y <= image->h) {
				DrawPartialImage(image, vanish_x, vanish_y, 0,
					vanish_y - pos_y, image->w, image->h - (vanish_y - pos_y));
			} else {
				vanishing = VANISHED;
			}
			break;
		case EAST:
			if (vanish_x - pos_x >= 0) {
				DrawPartialImage(image, pos_x, pos_y, 0, 0,
					vanish_x - pos_x, image->h);
			} else {
				vanishing = VANISHED;
			}
			break;
		case SOUTH:
			if (vanish_y - pos_y >= 0) {
				DrawPartialImage(image, pos_x, pos_y, 0, 0,
					image->w, vanish_y - pos_y);
			} else {
				vanishing = VANISHED;
			}
			break;
		case WEST:
			if (vanish_x - pos_x <= image->w) {
				DrawPartialImage(image, vanish_x, vanish_y, vanish_x - pos_x, 0,
					image->w - (vanish_x - pos_x), image->h);
			} else {
				vanishing = VANISHED;
			}
			break;
		}
		break;
	case APPEARING:
		/* In this case, the space being drawn is reversed, and upon completion
		 * the laser is reported PRESENT.
		 */
		switch(dir) {
		case NORTH:
			if (appear_y - pos_y <= image->h) {
				DrawPartialImage(image, pos_x, pos_y, 0, 0,
					image->w, appear_y - pos_y);
			} else {
				vanishing = PRESENT;
			}
			break;
		case EAST:
			if (appear_x - pos_x >= 0) {
				DrawPartialImage(image, appear_x, appear_y, appear_x - pos_x, 0,
					image->w - (appear_x - pos_x), image->h);
			} else {
				vanishing = PRESENT;
			}
			break;
		case SOUTH:
			if (appear_y - pos_y >= 0) {
				DrawPartialImage(image, appear_x, appear_y, 0,
					appear_y - pos_y, image->w, image->h - (appear_y - pos_y));
			} else {
				vanishing = PRESENT;
			}
			break;
		case WEST:
			if (appear_x - pos_x <= image->w) {
				DrawPartialImage(image, pos_x, pos_y, 0, 0,
					appear_x - pos_x, image->h);
			} else {
				vanishing = PRESENT;
			}
			break;
		}
		if (vanishing != PRESENT) break;
		/* Otherwise, we just changed to PRESENT but we haven't drawn it yet,
		 * so fall through to the PRESENT case.
		 */
	case PRESENT:
		this->Sprite::draw();
		break;
	}
}

void Laser::move() {
	this->Sprite::move();
	if (pos_x < -image->w || pos_x >= SCREEN_WIDTH ||
		pos_y < -image->h || pos_y >= SCREEN_HEIGHT) {
		this->vanish();
	}
}
