/***************************************************************************
 *   Copyright (C) 2009 by Ben Wolf   *
 *   bswolf@cs.cmu.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _SPRITE_H
#define _SPRITE_H

#include "graphics.h"
#include "direction.h"

/** A Sprite is a drawable object that appears on the playing screen.
 *  This class is intended to be a superclass of all objects that appear
 *  in a level, so that the level only needs to maintain a list of Sprites
 *  that should be drawn by the graphics engine.
 */
class Sprite {
private:
	int id;

protected:
	int pos_x;
	int pos_y;
	int vel_x;
	int vel_y;
	SDL_Surface* image;

public:
	/* Preload the images so that we don't have to load
	 * every time we create a Sprite.
	 */
	Sprite(int posx, int posy, SDL_Surface* image);

	/** @brief Moves the object in the specified direction.
	 *
	 *  Currently does not sanity check for running into a boundary
	 *  or wall, or running off the screen.
	 *
	 *  @param dir The direction in which to step the object.
	 */
	void step(int dir);

	/** @brief Moves the object in the direction specified by its current
	 *         velocity a distance specified by its current velocity.
	 */
	virtual void move();

	/** @brief Returns the current x-position of this Sprite.
	 */
	int getXPosition();

	/** @brief Returns the current y-position of this Sprite.
	 */
	int getYPosition();

	/** @brief Returns the current x-velocity of this Sprite.
	 */
	int getXVelocity();

	/** @brief Returns the current y-velocity of this Sprite.
	 */
	int getYVelocity();

	/** @brief Sets a new value for the x-velocity of this Sprite.
	 *
	 *  @param vel The new x-velocity.
	 */
	void setXVelocity(int vel);

	/** @brief Sets a new value for the y-velocity of this Sprite.
	 *
	 *  @param vel The new y-velocity.
	 */
	void setYVelocity(int vel);

	/** @brief Returns the image used for this Sprite.
	 */
	SDL_Surface* getImage();

	/** @brief Draws the image at its current location on screen.
	 *
	 */
	virtual void draw();
};

#endif /* _SPRITE_H */
