/***************************************************************************
 *   Copyright (C) 2009 by Ben Wolf   *
 *   bswolf@cs.cmu.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "graphics.h"
#include "Wall.h"
#include "debug.h"
#include <algorithm>
using namespace std;

static SDL_Surface *findImage(int x1, int y1, int x2, int y2) {
	if (x1 == x2) return GetImage(IMG_WALL_V);
	if (y1 == y2) return GetImage(IMG_WALL_H);
	int d = (x1-x2)*(y1-y2);
	if (d > 0) {
		// either x2 > x1 and y2 > y1 or x2 < x1 and y2 < y1
		return GetImage(IMG_WALL_UL);
	} else if (d < 0) {
		return GetImage(IMG_WALL_UR);
	}
	return GetImage(IMG_WALL2);
}

Wall::Wall(int x1, int y1, int x2, int y2) :
			Sprite(x1, y1, findImage(x1, y1, x2, y2)) {
	/* Enforce that (x1,y1) is the uppermost point. */
	if (y1 < y2 || (y1 == y2 && x1 < x2)) {
		this->x1 = x1;
		this->y1 = y1;
		this->x2 = x2;
		this->y2 = y2;
	} else {
		this->x1 = x2;
		this->y1 = y2;
		this->x2 = x1;
		this->y2 = y1;
	}

	if (x1 == x2) this->direction = NORTH;
	else if (y1 == y2) this->direction = EAST;
	else if ((x1-x2)*(y1-y2) > 0) this->direction = SOUTHEAST;
	else if ((x1-x2)*(y1-y2) < 0) this->direction = NORTHEAST;
	else {
		this->direction = -1;
	}
}

int Wall::getDirection() {
	return this->direction;
}

#define ABOVE 1
#define BELOW -1

bool checkLine(int x, int y, int x1, int y1, int x2, int y2, int side) {
	int eq = (y - y1) * (x2 - x1) - (y2 - y1) * (x - x1);
	return (side*eq > 0);
}

bool Wall::doesCollide(int x, int y) {
	/* These all assume point 1 is the UPPERMOST point
	 * (or LEFTMOST point if y1==y2) */
	switch(this->direction) {
	case NORTH:
	case SOUTH: return (x >= x1 && x <= x1 + this->image->w && y>=y1 && y<=y2);
	case EAST:
	case WEST: return (x >= x1 && x <= x2 && y>=y1 && y<=y1 + this->image->h);
	/* For the diagonal lines, check whether the point lies within a certain
	 * "thickness" of the center line, by checking that it's both above a line
	 * below the center line and below a line above the center line.
	 */
	case NORTHWEST:
	case SOUTHEAST: return (x >= x1 && x <= x2 && y >= y1 && y <= y2 &&
				/* Logically, we should be looking ABOVE this line and
				 * BELOW the other line... I don't know why this is instead
				 * the case.
				 */
				checkLine(x, y, x1, y1+THICKNESS, x2-THICKNESS, y2, BELOW) &&
				checkLine(x, y, x1+THICKNESS, y1, x2, y2-THICKNESS, ABOVE));
	case NORTHEAST:
	case SOUTHWEST: return (x >= x2 && x <= x1 && y >= y1 && y <= y2 &&
				checkLine(x, y, x1, y1+THICKNESS, x2+THICKNESS, y2, ABOVE) &&
				checkLine(x, y, x1-THICKNESS, y1, x2, y2-THICKNESS, BELOW));
	default: {
		db_log(DB_WALL, "Wall in bad direction!\n");
		return false;
		}
	}
}

void Wall::draw() {
	/* Again, assuming point 1 is the uppermost point... */
	switch(this->direction) {
	case NORTH:
	case SOUTH:
		for (int starth = y1, endh = y2 + image->w,
				diff = min(endh - starth, image->h);
				endh - starth > 0;
				starth += diff, diff = min(endh - starth, image->h)) {
			DrawPartialImage(image, x1, starth, 0, 0, image->w, diff);
		}
		break;
	case EAST:
	case WEST:
		for (int startw = x1, endw = x2 + image->h,
				diff = min(endw - startw, image->w);
				endw - startw > 0;
				startw += diff, diff = min(endw - startw, image->w)) {
			DrawPartialImage(image, startw, y1, 0, 0, diff, image->h);
		}
		break;
	case NORTHWEST:
	case SOUTHEAST:
		/* We only need to track one dimension, as both dimensions are equal. */
		for (int startw = x1, endw = x2 + THICKNESS,
				diff = min(endw - startw, image->w),
				starth = y1; endw - startw > THICKNESS;
				startw += diff - THICKNESS,	starth += diff - THICKNESS,
				diff = min(endw - startw, image->w)) {
			/* If we don't have a full length left, move back an amount
			 * so that we overlap with the previous length.
			 * PROBLEM: This doesn't allow for drawing diagonal walls shorter
			 * than the full length.
			 */
			DrawImage(image, startw - (image->w - diff),
				starth - (image->h - diff));
			//DrawPartialImage(image, startw, starth, 0, 0, diff, diff);
		}
		break;
	case NORTHEAST:
	case SOUTHWEST:
		/* We only need to track one dimension, as both dimensions are equal. */
		for (int starth = y1, endh = y2, diff = min(endh - starth, image->h),
				startw = x1; endh - starth > THICKNESS;
				startw -= diff - THICKNESS,	starth += diff - THICKNESS,
				diff = min(endh - starth, image->h)) {
			/* If we don't have a full length left, move back an amount
			 * so that we overlap with the previous length.
			 * PROBLEM: This doesn't allow for drawing diagonal walls shorter
			 * than the full length.
			 */
			DrawImage(image, startw - diff,
				starth - (image->h - diff));
			//DrawPartialImage(image, startw - diff, starth,
			//	image->w - diff, 0, diff, diff);
		}
		break;
	default:
		this->Sprite::draw();
	}
}
